import math
import torch
import torch.nn as nn
from typing import Optional

from e2eAIOK.DeNas.module.asr.attention import MultiheadAttention, PositionalwiseFeedForward
from e2eAIOK.DeNas.module.asr.normalization import LayerNorm

class TransformerDecoderLayer(nn.Module):
    def __init__(
        self,
        d_ffn,
        nhead,
        d_model,
        kdim=None,
        vdim=None,
        dropout=0.0,
        activation=nn.ReLU,
        normalize_before=False
    ):
        super().__init__()
        self.nhead = nhead

        self.self_attn = MultiheadAttention(
            nhead=nhead,
            d_model=d_model,
            kdim=kdim,
            vdim=vdim,
            dropout=dropout,
        )
        self.mutihead_attn = MultiheadAttention(
            nhead=nhead,
            d_model=d_model,
            kdim=kdim,
            vdim=vdim,
            dropout=dropout,
        )

        self.pos_ffn = PositionalwiseFeedForward(
            d_ffn=d_ffn,
            input_size=d_model,
            dropout=dropout,
            activation=activation,
        )

        # normalization layers
        self.norm1 = LayerNorm(d_model, eps=1e-6)
        self.norm2 = LayerNorm(d_model, eps=1e-6)
        self.norm3 = LayerNorm(d_model, eps=1e-6)
        self.dropout1 = torch.nn.Dropout(dropout)
        self.dropout2 = torch.nn.Dropout(dropout)
        self.dropout3 = torch.nn.Dropout(dropout)

        self.normalize_before = normalize_before

    def forward(
        self,
        tgt,
        memory,
        tgt_mask=None,
        memory_mask=None,
        tgt_key_padding_mask=None,
        memory_key_padding_mask=None,
        pos_embs_tgt=None,
        pos_embs_src=None,
    ):
        if self.normalize_before:
            tgt1 = self.norm1(tgt)
        else:
            tgt1 = tgt

        # self-attention over the target sequence
        tgt2, self_attn = self.self_attn(
            query=tgt1,
            key=tgt1,
            value=tgt1,
            attn_mask=tgt_mask,
            key_padding_mask=tgt_key_padding_mask,
            pos_embs=pos_embs_tgt,
        )

        # add & norm
        tgt = tgt + self.dropout1(tgt2)
        if not self.normalize_before:
            tgt = self.norm1(tgt)

        if self.normalize_before:
            tgt1 = self.norm2(tgt)
        else:
            tgt1 = tgt

        # multi-head attention over the target sequence and encoder states

        tgt2, multihead_attention = self.mutihead_attn(
            query=tgt1,
            key=memory,
            value=memory,
            attn_mask=memory_mask,
            key_padding_mask=memory_key_padding_mask,
            pos_embs=pos_embs_src,
        )

        # add & norm
        tgt = tgt + self.dropout2(tgt2)
        if not self.normalize_before:
            tgt = self.norm2(tgt)

        if self.normalize_before:
            tgt1 = self.norm3(tgt)
        else:
            tgt1 = tgt

        tgt2 = self.pos_ffn(tgt1)

        # add & norm
        tgt = tgt + self.dropout3(tgt2)
        if not self.normalize_before:
            tgt = self.norm3(tgt)

        return tgt, self_attn, multihead_attention


class TransformerDecoder(nn.Module):
    def __init__(
        self,
        num_layers,
        nhead,
        d_ffn,
        d_model,
        kdim=None,
        vdim=None,
        dropout=0.0,
        activation=nn.ReLU,
        normalize_before=False
    ):
        super().__init__()
        self.layers = torch.nn.ModuleList(
            [
                TransformerDecoderLayer(
                    d_ffn=d_ffn,
                    nhead=nhead,
                    d_model=d_model,
                    kdim=kdim,
                    vdim=vdim,
                    dropout=dropout,
                    activation=activation,
                    normalize_before=normalize_before,
                )
                for _ in range(num_layers)
            ]
        )
        self.norm = LayerNorm(d_model, eps=1e-6)

    def forward(
        self,
        tgt,
        memory,
        tgt_mask=None,
        memory_mask=None,
        tgt_key_padding_mask=None,
        memory_key_padding_mask=None,
        pos_embs_tgt=None,
        pos_embs_src=None,
    ):
        output = tgt
        self_attns, multihead_attns = [], []
        for dec_layer in self.layers:
            output, self_attn, multihead_attn = dec_layer(
                output,
                memory,
                tgt_mask=tgt_mask,
                memory_mask=memory_mask,
                tgt_key_padding_mask=tgt_key_padding_mask,
                memory_key_padding_mask=memory_key_padding_mask,
                pos_embs_tgt=pos_embs_tgt,
                pos_embs_src=pos_embs_src,
            )
            self_attns.append(self_attn)
            multihead_attns.append(multihead_attn)
        output = self.norm(output)

        return output, self_attns, multihead_attns


import torch
import torch.nn.functional as F

from e2eAIOK.DeNas.module.layernorm_base import LayerNormBase


class LayerNorm(LayerNormBase):
    """
    Applies layer normalization to the input tensor.
    """

    def __init__(
        self,
        input_size=None,
        input_shape=None,
        eps=1e-05,
        elementwise_affine=True,
    ):
        if input_shape is not None:
            input_size = input_shape[2:]
        super().__init__(input_size, eps, elementwise_affine)

    def forward(self, x):
        return F.layer_norm(
            x, self.normalized_shape, self.weight, self.bias, self.eps)



import sys
import time
import e2eAIOK.common.trainer.utils.utils as utils
import random
import torch
import logging
import e2eAIOK.common.trainer.utils.extend_distributed as ext_dist
from abc import ABC, abstractmethod
from e2eAIOK.common.trainer.data_builder import DataBuilder
from e2eAIOK.common.trainer.model_builder import ModelBuilder
 
class TorchTrainer():
    """
    The basic trainer class for all models

    Note:
        You should implement specfic model trainer under model folder like vit_trainer
    """
    def __init__(self, cfg, model, train_dataloader, eval_dataloader, optimizer, criterion, scheduler, metric):
        super().__init__()
        self.cfg = cfg
        self.model = model
        self.train_dataloader = train_dataloader
        self.eval_dataloader = eval_dataloader
        self.optimizer = optimizer
        self.criterion = criterion
        self.scheduler = scheduler
        self.metric = metric
        
    def _pre_process(self):
        """
            trainer pre process to prepare trainer environment
        """
        utils.init_log()
        self.logger = logging.getLogger('Trainer')
        self.logger.info(f"Trainer config: {self.cfg}")
        self._dist_wrapper()

    def _post_process(self):
        """
            trainer post process
        """
        self.logger.info("Trainer complete")
        
    def _dist_wrapper(self):
        """
            wrapper model for distributed training
        """
        if ext_dist.my_size > 1:
            self.model = ext_dist.DDP(self.model)
    
    def _is_early_stop(self, metric):
        """
            check whether training achieved pre-defined metric threshold
        """
        
        return metric >= self.cfg["metric_threshold"] if "metric_threshold" in self.cfg else False

    def train_one_epoch(self, epoch):
        """
            train one epoch
        """
        metric_logger = utils.MetricLogger(delimiter="  ")
        metric_logger.add_meter('lr', utils.SmoothedValue(window_size=1, fmt='{value:.6f}'))

        self.model.train()

        for inputs, targets in self.train_dataloader:
            outputs = self.model(inputs)
            loss = self.criterion(outputs, targets)

            loss_value = loss.item()
            self.optimizer.zero_grad()       
            loss.backward()
            self.optimizer.step()
        
            metric_logger.update(loss=loss_value)
            metric_logger.update(lr=self.optimizer.param_groups[0]["lr"])

            self.global_step += 1

            if 'eval_step' in self.cfg and self.global_step % self.cfg.eval_step == 0:
                self.model.eval()
                metric = utils.create_metric(self.cfg ,self.model, self.data_loader)

        # gather the stats from all processes
        metric_logger.synchronize_between_processes()
        self.logger.info("Averaged stats:", metric_logger)

    def evaluate(self, epoch):
        """
            evaluatiuon
        """
        metric_logger = utils.MetricLogger(delimiter="")

        self.model.eval()
        
        for inputs, target in self.eval_dataloader:
            output = self.model(inputs)
            loss = self.criterion(output, target)
            metric = self.metric(output, target)

            metric_logger.update(loss=loss.item())
            for k in metric.keys():
                metric_logger.meters[k].update(metric[k], n=self.cfg.eval_batch_size)
        # gather the stats from all processes
        metric_logger.synchronize_between_processes()
        output_str = []
        for name, meter in metric_logger.meters.items():
            output_str.append(
                "{}: {}".format(name, str(meter))
            )
        self.logger.info(output_str)

        if "output_dir" in self.cfg:
            checkpoint_paths = [self.cfg.output_dir + '/checkpoint.pth']
            for checkpoint_path in checkpoint_paths:
                utils.save_model({
                    'model': self.model.state_dict(),
                    'optimizer': self.optimizer.state_dict(),
                    'epoch': epoch,
                }, checkpoint_path)

    def fit(self):
        """
            trainint and evaluation
        """
        self._pre_process()
        start_time = time.time()
        for i in range(1, self.cfg.train_epochs+1):
            train_start = time.time()
            self.train_one_epoch(i)
            if i % self.cfg.eval_epochs == 0:
                eval_start = time.time()
                metric = self.evaluate(i)
                self.logger.info(F"Evaluate time:{time.time() - eval_start}")
                if self._is_early_stop(metric):
                    self.logger.info(f"Metric {metric} got threshold {self.cfg['metric_threshold']}, early stop")
                    break
            self.logger.info(F"Epoch {i} training time:{time.time() - train_start}")

        self.logger.info(F"Total time:{time.time() - start_time}")
        self._post_process()



import os
import sys
from setuptools import setup, find_packages
try:
    # pip >=20
    from pip._internal.network.session import PipSession
    from pip._internal.req import parse_requirements
except ImportError:
    try:
        # 10.0.0 <= pip <= 19.3.1
        from pip._internal.download import PipSession
        from pip._internal.req import parse_requirements
    except ImportError:
        # pip <= 9.0.3
        from pip.download import PipSession
        from pip.req import parse_requirements

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

e2eaiok_home = os.path.join(os.path.dirname(os.path.abspath(__file__)), "e2eAIOK")
VERSION = open(os.path.join(e2eaiok_home, "version"), 'r').read().strip()

def setup_package(args):
    metadata = dict(
        name=args["name"],
        version=VERSION,
        author="INTEL AIA BDF",
        author_email="chendi.xue@intel.com",
        license='Apache License',
        description="Intel® End-to-End AI Optimization Kit",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/intel/e2eAIOK",
        project_urls={
            "Bug Tracker": "https://github.com/intel/e2eAIOK/issues",
        },
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
        ],
        packages=args["packages"],
        package_data = args["package_data"],
        python_requires=">=3.6",
        zip_safe=False,
        install_requires=args["install_requires"]
    )
    setup(**metadata)

if __name__ == '__main__':
    args = dict(
        name = "e2eAIOK",
        packages = find_packages(exclude=["RecDP", "RecDP.*", "modelzoo", "example"]),
        install_requires = [],
        package_data = {}
    )
    if "--denas" in sys.argv:
        args["name"] = "e2eAIOK-denas"
        args["packages"] = find_packages(exclude=["RecDP", "RecDP.*", "modelzoo", "example", \
                                                    "e2eAIOK.SDA", "e2eAIOK.SDA.*", "e2eAIOK.dataloader", "e2eAIOK.utils",\
                                                    "e2eAIOK.ModelAdapter", "e2eAIOK.ModelAdapter.*"])
        args["package_data"] = {'e2eAIOK': ['version','common/default.conf']},
        install_reqs = parse_requirements("e2eAIOK/DeNas/requirements.txt", session=False)
        # handle pip version compatibility
        try:
            args["install_requires"] = [str(ir.req) for ir in install_reqs]
        except AttributeError:
            args["install_requires"] = [str(ir.requirement) for ir in install_reqs]        
        sys.argv.remove("--denas")
    elif "--sda" in sys.argv:
        args["name"] = "e2eAIOK-sda"
        args["packages"] = find_packages(exclude=["RecDP", "RecDP.*", "modelzoo", "example", \
                                                    "e2eAIOK.DeNas", "e2eAIOK.DeNas.*",\
                                                    "e2eAIOK.ModelAdapter", "e2eAIOK.ModelAdapter.*"])
        args["package_data"] = {'e2eAIOK': ['version','common/default.conf']},
        install_reqs = parse_requirements("e2eAIOK/SDA/requirements.txt", session=False)
        # handle pip version compatibility
        try:
            args["install_requires"] = [str(ir.req) for ir in install_reqs]
        except AttributeError:
            args["install_requires"] = [str(ir.requirement) for ir in install_reqs] 
        sys.argv.remove("--sda")
    elif "--ma" in sys.argv:
        args["name"] = "e2eAIOK-ModelAdapter"
        args["packages"] = find_packages(exclude=["RecDP", "RecDP.*", "modelzoo", "example", \
                                                "e2eAIOK.SDA", "e2eAIOK.SDA.*", "e2eAIOK.dataloader", "e2eAIOK.utils",\
                                                "e2eAIOK.DeNas", "e2eAIOK.DeNas.*"])
        args["package_data"] = {'e2eAIOK': ['version','common/default.conf','ModelAdapter/default_ma.conf']}
        install_reqs = parse_requirements("e2eAIOK/ModelAdapter/requirements.txt", session=False)
        # handle pip version compatibility
        try:
            args["install_requires"] = [str(ir.req) for ir in install_reqs]
        except AttributeError:
            args["install_requires"] = [str(ir.requirement) for ir in install_reqs] 
        sys.argv.remove("--ma")
    else:
        args["package_data"] = {'e2eAIOK': ['version','common/default.conf','ModelAdapter/default_ma.conf']}
    setup_package(args)



import subprocess
import yaml
import logging
import time

from e2eAIOK.common.utils import *
from e2eAIOK.SDA.modeladvisor.BaseModelAdvisor import BaseModelAdvisor

class TwitterRecSysAdvisor(BaseModelAdvisor):
    def __init__(self, dataset_meta_path, train_path, eval_path, settings):
        super().__init__(dataset_meta_path, train_path, eval_path, settings)
        logging.basicConfig(level = logging.INFO,format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger('sigopt')

        self.train_path = train_path
        self.test_path = eval_path
        self.dataset_meta_path = dataset_meta_path
        self.saved_path = self.params['save_path']
        self.train_python = "/opt/intel/oneapi/intelpython/latest/bin/python"
        self.train_script = "/home/vmagent/app/e2eaiok/modelzoo/TwitterRecSys2021/model_e2eaiok/xgboost/train.py"

    def update_metrics(self):
        result_metrics_path = os.path.join(self.params['model_saved_path'],
                                           "result.yaml")
        if not os.path.exists(result_metrics_path):
            raise FileNotFoundError(
                f"{self.train_script} completed, while we can't find \
                    result {result_metrics_path} file.")
        with open(result_metrics_path) as f:
            results = yaml.load(f, Loader=yaml.FullLoader)
        metrics = []
        metrics.append({'name': 'AP', 'value': results['AP']})
        metrics.append({'name': 'RCE', 'value': results['RCE']})

        self.params['model_parameter']['metrics'] = metrics
        return self.params['model_parameter']['metrics']

    def initialize_model_parameter(self, assignments = None):
        config = {}
        tuned_parameters = {}
        if assignments:
            tuned_parameters['max_depth'] = assignments['max_depth']
            tuned_parameters['learning_rate'] = assignments['learning_rate']
            tuned_parameters['subsample'] = assignments['subsample']
            tuned_parameters['colsample_bytree'] = assignments['colsample_bytree']
            tuned_parameters['num_boost_round'] = assignments['num_boost_round']
        else:
            tuned_parameters['max_depth'] = 8
            tuned_parameters['learning_rate'] = float(0.1)
            tuned_parameters['subsample'] = float(0.8)
            tuned_parameters['colsample_bytree'] = float(0.8)
            tuned_parameters['num_boost_round'] = 250

        config['tuned_parameters'] = tuned_parameters
        self.params['model_parameter'] = config
        self.params['model_saved_path'] = os.path.join(self.params['save_path'], get_hash_string(str(tuned_parameters)))
        return config

    def generate_sigopt_yaml(self, file='RecSys_sigopt.yaml'):
        config = {}
        config['project'] = 'e2eaiok'
        config['experiment'] = 'RecSys'
        parameters = [{'name': 'max_depth', 'bounds': {'min': 5, 'max': 20}, 'type': 'int'},
                    {'name': 'learning_rate', 'bounds': {'min': 0.0, 'max': 1.0}, 'type': 'double'},
                    {'name': 'subsample', 'bounds': {'min': 0.5, 'max': 1}, 'type': 'double'},
                    {'name': 'colsample_bytree', 'bounds': {'min': 0.5, 'max': 1.0}, 'type': 'double'},
                    {'name': 'num_boost_round', 'bounds': {'min': 100, 'max': 1000}, 'type': 'int'}]
        user_defined_parameter = self.params['model_parameter']['parameters'] if ('model_parameter' in self.params) and ('parameters' in self.params['model_parameter']) else None
        config['parameters'] = parameters
        if user_defined_parameter:
            self.logger.info(f"Update with user defined parameters {user_defined_parameter}")
            update_list(config['parameters'], user_defined_parameter)
        
        config['metrics'] = [
            {'name': 'AP', 'strategy': 'optimize', 'objective': 'maximize'},
            {'name': 'RCE', 'objective': 'maximize'}
        ]
        user_defined_metrics = self.params['model_parameter']['metrics'] if ('model_parameter' in self.params) and ('metrics' in self.params['model_parameter']) else None
        if user_defined_metrics:
            self.logger.info(f"Update with user defined parameters {user_defined_metrics}")
            update_list(config['metrics'], user_defined_metrics)
        config['observation_budget'] = self.params['observation_budget']

        # save to local disk
        saved_path = os.path.join(self.params['save_path'], file)
        with open(saved_path, 'w') as f:
            yaml.dump(config, f)
        return config

    def train_model(self, args):
        start_time = time.time()
        self.launch(args)
        self.training_time = time.time() - start_time
        metrics = self.update_metrics()
        return self.training_time, args['model_saved_path'], metrics
    
    def launch(self, args):
        cmd = f"{self.train_python} -u "
        cmd += f"{self.train_script} "
        cmd += f"--stage {args['stage']} "
        cmd += f"--target {args['target']} "
        cmd += f"--train_data_path {self.train_path} "
        cmd += f"--valid_data_path {self.test_path} "
        cmd += f"--model_save_path {args['model_saved_path']} "
        cmd += f"--max_depth {args['model_parameter']['tuned_parameters']['max_depth']} "
        cmd += f"--learning_rate {args['model_parameter']['tuned_parameters']['learning_rate']} "
        cmd += f"--subsample {args['model_parameter']['tuned_parameters']['subsample']} "
        cmd += f"--colsample_bytree {args['model_parameter']['tuned_parameters']['colsample_bytree']} "
        cmd += f"--num_boost_round {args['model_parameter']['tuned_parameters']['num_boost_round']} "
        self.logger.info(f'training launch command: {cmd}')

        process = subprocess.Popen(cmd, shell=True)
        process.wait()




#!/usr/bin/python
# -*- coding: utf-8 -*-
from e2eAIOK.common.trainer.model.cv.resnet_cifar import ResNet18 as resnet18_cifar
from e2eAIOK.common.trainer.model.cv.resnet_cifar import ResNet50 as resnet50_cifar
from e2eAIOK.common.trainer.model.cv.lenet import LeNet
from e2eAIOK.common.trainer.model.model_utils.model_utils import initWeights
from e2eAIOK.common.trainer.model.model_builder_cv import ModelBuilderCV
import logging
import torch
import timm
import sys, os
import argparse

def createBackbone(cfg, model_type, num_classes, initial_pretrain=False, pretrain=""):
    ''' create backbone by name 
    :param cfg: configurations
    :param num_classes: number of classes
    :param initial_pretrain: whether load default pretrain weights
    :pretrain: pretrain model path
    '''
    backbone_name = model_type.lower()
    if backbone_name == 'lenet':
        model = LeNet(num_classes)#.cuda()
    elif backbone_name == "resnet18_cifar":
        model = resnet18_cifar(num_classes=num_classes)
    elif backbone_name == "resnet50_cifar":
        model = resnet50_cifar(num_classes=num_classes)
    elif backbone_name == 'resnet18':
        model = timm.create_model('resnet18', pretrained=initial_pretrain, num_classes=num_classes)
    elif backbone_name == 'resnet50':
        model = timm.create_model('resnet50', pretrained=initial_pretrain, num_classes=num_classes)
    elif backbone_name == 'mobilenet_v3':
        model = timm.create_model('mobilenetv3_large_100', pretrained=initial_pretrain, num_classes=num_classes)
    elif backbone_name == 'vit_base':
        model = timm.create_model('vit_base_patch16_224_miil', pretrained=initial_pretrain, num_classes=num_classes)
    elif backbone_name == "denas_cnn":
        curdir = os.path.abspath(os.path.dirname(__file__))
        sys.path.append(os.path.join(curdir,"DeNas"))
        from .DeNas.trainer.model.cv.cnn_model_builder import CNNModelBuilder
        args_dict = {
            "num_classes": num_classes,
            "best_model_structure":os.path.join(curdir,"DeNas/best_model_structure.txt")
        }
        args = argparse.Namespace(**args_dict)
        model_builder = CNNModelBuilder(args)
        with open(args.best_model_structure, 'r') as f:
            arch = f.readlines()[-1]
        model = model_builder.create_model(arch)
    elif backbone_name == "huggingface_vit_base_224_in21k_ft_cifar100":
        from transformers import ViTForImageClassification
        model = ViTForImageClassification.from_pretrained('edumunozsala/vit_base-224-in21k-ft-cifar100')
    elif backbone_name == "unet":
        from .unet.generic_UNet_DA import Generic_UNet_DA
        args_dict = {
            "threeD": True, 
            "input_channels": 1, 
            "base_num_features": 30, 
            "num_classes": num_classes, 
            "num_conv_per_stage": 2,  
            "pool_op_kernel_sizes": None,
            "conv_kernel_sizes": None
        }
        model = Generic_UNet_DA(**args_dict)
    else:
        logging.error("[%s] is not supported"%backbone_name)
        raise NotImplementedError("[%s] is not supported"%backbone_name)

    if (not initial_pretrain) and (not pretrain):
        model.apply(initWeights)

    modelbuilder = ModelBuilderCV(cfg, model)
    model = modelbuilder.create_model(pretrain)
       
    return model


#!/usr/bin/python
# -*- coding: utf-8 -*-
import torch.nn as nn
import torch
import logging
from .grl import GradientReverseLayer

class AdversarialAdapter(nn.Module):
    ''' AdversarialNetwork used for adversarial domain adaption

    '''
    def __init__(self, in_feature, hidden_size,dropout_rate,grl_coeff_alpha,grl_coeff_high,max_iter):
        ''' init method.

        :param in_feature: adversarial network input
        :param hidden_size: hidden size
        :param dropout_rate: dropout rate
        :param grl_coeff_alpha: GradientReverseLayer alpha argument to calculate coef
        :param grl_coeff_high: GradientReverseLayer coef high
        :param max_iter: max iter for one epoch
        '''
        super(AdversarialAdapter, self).__init__()
        self.ad_layer1 = nn.Linear(in_feature, hidden_size)
        self.ad_layer2 = nn.Linear(hidden_size, hidden_size//2)
        self.ad_layer3 = nn.Linear(hidden_size//2, 1)
        self.relu1 = nn.ReLU()
        self.relu2 = nn.ReLU()
        self.dropout1 = nn.Dropout(dropout_rate)
        self.dropout2 = nn.Dropout(dropout_rate)
        self.sigmoid = nn.Sigmoid()
        self.grl = GradientReverseLayer(coeff_alpha=grl_coeff_alpha,coeff_high=grl_coeff_high,
                                        max_iter=max_iter,enable_step=self.training)
        self.iter_num = 0

    def forward(self, x,**kwargs):
        x = self.grl(x)
        # coeff = np.float(2.0 * self.grl_coeff_high / (1.0 + np.exp(-self.grl_coeff_alpha * self.iter_num / self.max_iter)) - self.grl_coeff_high)
        # if self.iter_num % 10 ==0:
        #     print(self.grl_coeff_alpha,self.grl_coeff_high,self.iter_num ,self.max_iter,coeff)
        # x.register_hook(lambda g: -g*coeff)
        x = self.ad_layer1(x)
        x = self.relu1(x)
        x = self.dropout1(x)
        x = self.ad_layer2(x)
        x = self.relu2(x)
        x = self.dropout2(x)
        x = self.ad_layer3(x)
        y = self.sigmoid(x)

        if self.training:
            self.iter_num += 1
        return y

    @classmethod
    def make_label(cls,shape,is_source):
        ''' create adapter label. If is_source = True then label is 1, else 0.

        :param shape: label shape
        :param is_source: whether it is source
        :return: adapter label
        '''
        if is_source:
            return torch.ones(shape,dtype=torch.float)   #.cuda()
        else:
            return torch.zeros(shape, dtype=torch.float)  # .cuda()

    def loss(self,output_prob,label,**kwargs):
        ''' adapter loss function

        :param output_prob: output probability
        :param label: ground truth
        :param kwargs: kwargs
        :return: loss
        '''
        raise NotImplementedError("must implement loss function")


# Todo: Add common nlp model features into this module in the future
import torch
from e2eAIOK.common.trainer.model_builder import ModelBuilder

class ModelBuilderNLP(ModelBuilder):
    def init(self, cfg):
        super().init(cfg)



import torch
import logging
from operator import itemgetter
from torch.utils.data import (
    RandomSampler,
    DistributedSampler,
)


logger = logging.getLogger(__name__)


class ReproducibleRandomSampler(RandomSampler):
    """A modification of RandomSampler which always returns the same values.

    Also look at `torch.utils.data.RandomSampler`. This has mostly
    the same behaviour and arguments, except for adding 'seed' and 'epoch' and
    not supporting 'generator'.

    Note
    ----
    Call `set_epoch` before every epoch. Otherwise, the sampler will produce the
    same sequence of indices every epoch.

    Arguments
    ---------
    data_source : Dataset
        The data source to sample indices for.
    seed : int
        The base seed to use for the random number generator. It is recommended
        to use a value which has a good mix of 0 and 1 bits.
    epoch : int
        The epoch to start at.
    """

    def __init__(self, data_source, seed=563375142, epoch=0, **kwargs):
        if "generator" in kwargs:
            MSG = (
                "Cannot give a separate generator when using "
                + "ReproducibleRandomSampler"
            )
            raise ValueError(MSG)
        super().__init__(data_source, **kwargs)
        self.seed = int(seed)
        self.epoch = epoch
        self.generator = torch.Generator()

    def set_epoch(self, epoch):
        """
        You can also just access self.epoch, but we maintain this interface
        to mirror torch.utils.data.distributed.DistributedSampler
        """
        self.epoch = epoch

    def __iter__(self):
        self.generator.manual_seed(self.seed + self.epoch)
        return super().__iter__()


class DistributedSamplerWrapper(DistributedSampler):
    """This wrapper allows using any sampler with Distributed Data Parallel (DDP) correctly.

    Passing blindly the sampler to each DDP process will cause to have access
    within each process to all the data in the dataset instead of only a subset
    of it which is unique to each process.  This wrapper prevents this and
    allows to use only a subset of the original data for each process.
    """

    def __init__(self, sampler, *args, **kwargs):
        super().__init__(dataset=sampler, *args, **kwargs)
        self.sampler = sampler

    def __iter__(self):
        # It is easiest to use a random access interface to the wrapped
        # sampler's indices, so we just fetch all indices from the wrapped
        # sampler
        sampler_indices = list(self.sampler.__iter__())
        indices_of_indices = super().__iter__()
        # Itemgetter fetches the wrapped sampler indices from the positions
        # pointed to by DistributedSampler
        return iter(itemgetter(*indices_of_indices)(sampler_indices))

    def set_epoch(self, epoch):
        """Pass set_epoch() through to DistributedSampler and the wrapper one"""
        super().set_epoch(epoch)
        if hasattr(self.sampler, "set_epoch"):
            self.sampler.set_epoch(epoch)



import os, sys, time, math
import torch
from torch import nn
import numpy as np
import gc
import torch
# TODO: separate domain specific ops
from e2eAIOK.DeNas.module.cv.Linear_super import LinearSuper
from e2eAIOK.DeNas.module.cv.layernorm_super import LayerNormSuper
from e2eAIOK.DeNas.module.cv.multihead_super import AttentionSuper
from e2eAIOK.DeNas.module.cv.embedding_super import PatchembedSuper
from e2eAIOK.DeNas.cv.supernet_transformer import TransformerEncoderLayer
from e2eAIOK.DeNas.cv.benchmark_network_latency import get_model_latency
from e2eAIOK.DeNas.nlp.supernet_bert import SuperBertEncoder
from e2eAIOK.DeNas.nlp.utils import get_bert_latency
from e2eAIOK.DeNas.module.asr.encoder import TransformerEncoder
from e2eAIOK.DeNas.module.asr.attention import MultiheadAttention
from e2eAIOK.DeNas.module.asr.linear import Linear

from torchsummary import summary

def network_weight_gaussian_init(net, model_type):
    with torch.no_grad():
        if model_type == "transformer":   
            for m in net.modules():
                if isinstance(m, LinearSuper):
                    nn.init.normal_(m.weight)
                    if hasattr(m, 'bias') and m.bias is not None:
                        nn.init.zeros_(m.bias)
                elif isinstance(m, TransformerEncoderLayer):
                    nn.init.ones_(m.fc1.weight)
                    nn.init.zeros_(m.fc1.bias)
                    nn.init.ones_(m.fc2.weight)
                    nn.init.zeros_(m.fc2.bias)
                    nn.init.ones_(m.attn_layer_norm.weight)
                    nn.init.zeros_(m.attn_layer_norm.bias)
                    nn.init.ones_(m.ffn_layer_norm.weight)
                    nn.init.zeros_(m.ffn_layer_norm.bias)
                elif isinstance(m, LayerNormSuper):
                    nn.init.normal_(m.weight)
                    if hasattr(m, 'bias') and m.bias is not None:
                        nn.init.zeros_(m.bias)
                elif isinstance(m, PatchembedSuper):
                    nn.init.normal_(m.proj.weight)
                    if hasattr(m.proj, 'bias') and m.proj.bias is not None:
                        nn.init.zeros_(m.proj.bias)
                else:
                    continue
    return net

def get_linear_layer_metric_array(model_type, net, metric):
    metric_array = []
    for layer in net.modules():
        if model_type == "transformer":
            if isinstance(layer, LinearSuper):
                metric_array.append(metric(layer))
            elif isinstance(layer, TransformerEncoderLayer):
                metric_array.append(metric(layer.fc1))
                metric_array.append(metric(layer.fc2))
        elif model_type == "bert":
            if isinstance(layer, SuperBertEncoder):
                for sub_layer in layer.layers:
                    metric_array.append(metric(sub_layer.intermediate.dense))
                    metric_array.append(metric(sub_layer.output.dense))
        elif model_type == "asr":
            if isinstance(layer, TransformerEncoder):
                for sub_layer in layer.layers:
                    metric_array.append(metric(sub_layer.pos_ffn.fc1))
                    metric_array.append(metric(sub_layer.pos_ffn.fc2))
    return metric_array

def get_attn_layer_metric_array(model_type, net, metric):
    def func(weight):
        if weight.grad is not None:
            return torch.norm(weight) * torch.norm(weight.grad)
        else:
            return torch.zeros_like(weight)
    score = 0 
    metric_array = []
    for layer in net.modules():
        if model_type == "transformer":
            if isinstance(layer, TransformerEncoderLayer):
                metric_array.append(metric(layer.attn.qkv))
                metric_array.append(metric(layer.attn.proj))
        elif model_type == "bert":
            if isinstance(layer, SuperBertEncoder):
                for sub_layer in layer.layers:
                    metric_array.append(metric(sub_layer.attention.self.query))
                    metric_array.append(metric(sub_layer.attention.self.key))
                    metric_array.append(metric(sub_layer.attention.self.value))
                    metric_array.append(metric(sub_layer.output.dense))
        elif model_type == 'asr':
            if isinstance(layer, TransformerEncoder):
                for sub_layer in layer.layers:
                    metric_array.append(func(sub_layer.self_att.att.in_proj_weight))
                    metric_array.append(metric(sub_layer.self_att.att.out_proj))
    return metric_array


def compute_diversity_score(model_type, net, *inputs):

    # Compute gradients with input of 1s
  
    net.zero_grad()
    if model_type == "transformer":
        input_dim = list(inputs[0][0, :].shape)
        inputs = torch.ones([1] + input_dim)
        output = net.forward(inputs)
    elif model_type == "bert":
        input_ids, input_masks, input_segments = inputs
        output, pooled_output = net.forward(input_ids, input_masks, input_segments)
    elif model_type == "asr":
        output, _ = net.encode(inputs[0])
    torch.sum(output).backward()

    # select the gradients that we want to use for search/prune
    def nuc_norm(layer):
        if layer.weight.grad is not None:
            return torch.norm(layer.weight) * torch.norm(layer.weight.grad)
        else:
            return torch.zeros_like(layer.weight)

    disversity_score_list = get_attn_layer_metric_array(model_type, net, nuc_norm)
    
    return disversity_score_list


def compute_saliency_score(model_type, net, *inputs):
    device = inputs[0].device

    # convert params to their abs. Keep sign for converting it back.
    @torch.no_grad()
    def linearize(net):
        signs = {}
        for name, param in net.state_dict().items():
            signs[name] = torch.sign(param)
            param.abs_()
        return signs

    # convert to orig values
    @torch.no_grad()
    def nonlinearize(net, signs):
        for name, param in net.state_dict().items():
            if 'weight_mask' not in name:
                param.mul_(signs[name])

    # keep signs of all params
    signs = linearize(net)
    net.zero_grad()

    if model_type == "transformer":
    # Compute gradients with input of 1s
        input_dim = list(inputs[0][0, :].shape)
        inputs = torch.ones([1] + input_dim)
        output = net.forward(inputs)
    elif model_type == "bert":
        input_ids, input_masks, input_segments = inputs
        output, pooled_output = net.forward(input_ids, input_masks, input_segments)
    elif model_type == "asr":
        output, _ = net.encode(inputs[0])

    torch.sum(output).backward()

    # select the gradients that we want to use for search/prune
    def synflow(layer):
        if layer.weight.grad is not None:
            return torch.abs(layer.weight * layer.weight.grad)
        else:
            return torch.zeros_like(layer.weight)

    grads_abs = get_linear_layer_metric_array(model_type, net, synflow)

    # apply signs of all params
    nonlinearize(net, signs)

    return grads_abs

def do_compute_nas_score_transformer(model_type, model, resolution, batch_size, mixup_gamma, subconfig=None, expressivity_weight=0, complexity_weight=0, diversity_weight=0, saliency_weight=0, latency_weight=0):
    
    expressivity_score = 0
    complexity_score = 0
    network_weight_gaussian_init(model,model_type)
    if subconfig is not None:
        model.module.set_sample_config(subconfig) if hasattr(model, 'module') \
            else model.set_sample_config(subconfig)
    model.train()
    model.requires_grad_(True)
    model.zero_grad()
    

    if model_type == "transformer":
        dtype = torch.float32
        input = torch.randn(size=[batch_size, 3, resolution, resolution],  dtype=dtype)
        disversity_score_list = compute_diversity_score(model_type, model, input)
    elif model_type == "bert":
        
        max_seq_length = resolution
        input_ids = [[9333-id] * max_seq_length for id in range(batch_size)]
        input_masks = max_seq_length * [1]
        input_segments = max_seq_length * [0]
        input_ids = torch.tensor(input_ids, dtype=torch.long)
        input_masks = torch.tensor([input_masks]*batch_size, dtype=torch.long)
        input_segments = torch.tensor([input_segments]*batch_size, dtype=torch.long)
        disversity_score_list = compute_diversity_score(model_type, model, input_ids, input_masks, input_segments)
    elif model_type == "asr":
        input = torch.randn(size=[batch_size, 400, 20, 64])
        disversity_score_list = compute_diversity_score(model_type, model, input)
    disversity_score = 0
    for grad_abs in disversity_score_list:
        if len(grad_abs.shape) == 0:
            disversity_score += grad_abs
        elif len(grad_abs.shape) == 2:
            disversity_score += float(torch.mean(torch.sum(grad_abs, dim=[1])))

    if model_type == "transformer":
        grads_abs_list = compute_saliency_score(model_type, model, input)
    elif model_type == "bert":
        grads_abs_list = compute_saliency_score(model_type, model, input_ids, input_masks, input_segments)
    elif model_type == "asr":
        grads_abs_list = compute_saliency_score(model_type, model, input)
   
    saliency_score = 0
    for grad_abs in grads_abs_list:
        if len(grad_abs.shape) == 4:
            saliency_score += float(torch.mean(torch.sum(grad_abs, dim=[1,2,3])))
        elif len(grad_abs.shape) == 2:
            saliency_score += float(torch.mean(torch.sum(grad_abs, dim=[1])))
        else:
            raise RuntimeError('only support grad shape of 4 or 2')
    if model_type == "transformer":
        latency = get_model_latency(model=model, batch_size=batch_size,
                                                        resolution=resolution,
                                                        in_channels=3, gpu=None, repeat_times=3,
                                                        fp16=False)    
    elif model_type == "bert":
        latency = get_bert_latency(model=model, batch_size=batch_size, max_seq_length=resolution, gpu=None, infer_cnt=10.)
    else:
        latency = 0
    score = (expressivity_score*expressivity_weight 
                    + complexity_score*complexity_weight 
                    + disversity_score*diversity_weight 
                    + saliency_score*saliency_weight)
    nas_score = score/(1 + latency*latency_weight)
    return nas_score, score, latency



import collections
import uuid


class CircularDependencyError(ValueError):
    """
    An error caused by running into circular dependencies while searching for
    an evaluation order in a DependencyGraph.
    """

    pass


DGNode = collections.namedtuple("DGNode", ["key", "edges", "data"])
# A node in DependencyGraph.


class DependencyGraph:
    """General-purpose dependency graph.

    Essentially a directed acyclic graph.
    Usually used to find an evaluation order for e.g. variable substitution
    The relation that an edge between A and B represents is:
    "A depends on B, i.e. B should be evaluated before A"

    Nodes can be added explicitly or they can be created implicitly
    while adding edges.
    Nodes have keys, which should be some hashable value that identifies
    the elements the graph represents in your use case. E.G. they can just
    be the variable name you want to substitute.
    However, if needed, more generally you can attach any data to a node
    (e.g. a path in your tree), and if so desired, a unique key can be
    created for you. You'll only need to know that key while adding edges
    to/from it.
    Implicit keys and explicit keys can also be mixed.
    """

    def __init__(self):
        self.digraph = []
        self.key2ind = {}
        # Guard for manual duplicates (but not implicitly added ones)
        self._manually_added_keys = []

    @staticmethod
    def get_unique_key():
        """Returns a unique hashable identifier."""
        return uuid.uuid4()

    def add_node(self, key=None, data=None):
        """Adds a node explicitly.

        Arguments
        ---------
        key : hashable, optional
            If not given, a key is created for you.
        data : Any, optional
            Any additional data you wish to attach to this node.

        Returns
        -------
        hashable
            The key that was used (either yours or generated).

        Raises
        ------
        ValueError
            If node with the given key has already been added explicitly
            (with this method, not "add_edge").
        """
        if key is None:
            key = self.get_unique_key()
        elif key in self._manually_added_keys:
            raise ValueError("Adding duplicate node: {key}".format(key=key))
        else:
            self._manually_added_keys.append(key)
        if key in self.key2ind:  # Implicitly added already; don't add again.
            ind = self.key2ind[key]
            node = self.digraph[ind]
            # All that this operation can do is add data:
            self.digraph[ind] = DGNode(node.key, node.edges, data)
            return key
        self.key2ind[key] = len(self.digraph)
        self.digraph.append(DGNode(key, [], data))
        return key

    def add_edge(self, from_key, to_key):
        """Adds an edge, and implicitly also creates nodes for keys which have
        not been seen before. This will not let you add data to your nodes.
        The relation encodes: "from_key depends on to_key"
        (to_key must be evaluated before from_key).

        Arguments
        ---------
        from_key : hashable
            The key which depends on.
        to_key : hashable
            The key which is depended on.

        Returns
        -------
        None
        """
        from_ind = self._get_ind_and_add_if_new(from_key)
        to_ind = self._get_ind_and_add_if_new(to_key)
        edges_list = self.digraph[from_ind].edges
        if to_ind not in edges_list:
            edges_list.append(to_ind)

    def _get_ind_and_add_if_new(self, key):
        # Used internally to implicitly add nodes for unseen keys
        if key not in self.key2ind:
            self.key2ind[key] = len(self.digraph)
            self.digraph.append(DGNode(key, [], None))
        return self.key2ind[key]

    def is_valid(self):
        """Checks if an evaluation order can be found.

        A dependency graph is evaluatable if there are no circular
        dependencies, i.e., the graph is acyclic.

        Returns
        -------
        bool
            Indicating if the graph is evaluatable.
        """
        return not self._find_first_cycle()

    def get_evaluation_order(self, selected_keys=None):
        """Finds one valid evaluation order.

        There can be many different valid
        orders.
        NOTE: Generates output one DGNode at a time. May generate DGNodes
        before it finds a circular dependency. If you really need to know
        whether an order can be found, check is_valid() first. However,
        the algorithm for finding cycles is essentially the same as the one
        used for finding an evaluation order, so for very large graphs...
        Ah well, but maybe then you should be using some other solution
        anyway.

        Arguments
        ---------
        selected_keys : list, None
            List of keys. If not None, only the selected keys are guaranteed
            in the evaluation order (along with the keys they depend on).

        Yields
        ------
        DGNode
            The added DGNodes in a valid evaluation order.
            See the DGNode namedtuple above.

        Raises
        ------
        CircularDependencyError
            If a circular dependency is found.
        """
        seen_ever = set()

        def toposort(root_ind, visited):
            """Implementation of topsort."""
            nonlocal seen_ever
            here = visited + [root_ind]
            if root_ind in visited:
                raise CircularDependencyError(
                    "{cycle}".format(
                        cycle=" -> ".join(
                            str(self.digraph[i].key) for i in here
                        )
                    )
                )
            if root_ind in seen_ever:
                return  # Yield nothing
            seen_ever = seen_ever.union(set([root_ind]))
            for to_ind in self.digraph[root_ind].edges:
                for ind in toposort(to_ind, visited=here):
                    yield ind
            yield root_ind

        if selected_keys is None:
            start_inds = range(len(self.digraph))
        else:
            start_inds = [self.key2ind[key] for key in selected_keys]

        for start_ind in start_inds:
            for ind in toposort(start_ind, []):
                yield self.digraph[ind]

    def _find_first_cycle(self):
        """Depth-first search based algorithm for finding cycles in the graph."""
        seen_ever = set()

        def cycle_dfs(root_ind, visited):
            """Implementation of cycle_dfs."""
            nonlocal seen_ever
            print(root_ind, visited)
            here = visited + [root_ind]
            if root_ind in visited:
                return here
            if root_ind in seen_ever:
                return []
            seen_ever = seen_ever.union(set([root_ind]))
            for to_ind in self.digraph[root_ind].edges:
                cycle = cycle_dfs(to_ind, here)
                if cycle:
                    return cycle
            return []

        for ind in range(len(self.digraph)):
            if ind not in seen_ever:
                cycle = cycle_dfs(ind, [])
                if cycle:
                    return cycle
        return []

    def __contains__(self, key):
        # Allows the syntax:
        # 'key' in dependency_graph
        return key in self.key2ind



import torch

from e2eAIOK.DeNas.asr.supernet_asr import TransformerASRSuper


def gen_transformer(
    input_size=1280, output_neurons=5000, d_model=512, encoder_heads=[4]*12,
    nhead=4, num_encoder_layers=12, num_decoder_layers=6, mlp_ratio=[4.0]*12, 
    d_ffn=2048, transformer_dropout=0.1
):
    model = TransformerASRSuper(
        input_size = input_size,
        tgt_vocab = output_neurons,
        d_model = d_model,
        encoder_heads = encoder_heads,
        nhead = nhead,
        num_encoder_layers = num_encoder_layers,
        num_decoder_layers = num_decoder_layers,
        mlp_ratio = mlp_ratio,
        d_ffn = d_ffn,
        dropout = transformer_dropout,
        activation = torch.nn.GELU,
        normalize_before = True
    )
    return model


import torch
from torch.utils.data import DataLoader
from torch.utils.data import IterableDataset
import logging
from torch.utils.data import DistributedSampler
import torch.distributed as dist

from e2eAIOK.DeNas.asr.data.dataio.batch import PaddedBatch
from e2eAIOK.DeNas.asr.data.dataio.dataset import DynamicItemDataset
from e2eAIOK.DeNas.asr.data.dataio.sampler import ReproducibleRandomSampler, DistributedSamplerWrapper
from e2eAIOK.DeNas.asr.utils.checkpoints import (
    mark_as_saver,
    mark_as_loader,
)


logger = logging.getLogger(__name__)

def get_dataloader(dataset, batch_size, distributed):
    if distributed:
        sampler = DistributedSampler(dataset, num_replicas=dist.get_world_size(), rank=dist.get_rank())
    else:
        sampler = None
    dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=(not distributed), collate_fn=PaddedBatch, drop_last=True, sampler=sampler)
    return dataloader

def make_dataloader(dataset, stage, dist, **loader_kwargs):
    if stage == 'train':
        loader_kwargs = train_loader_specifics(dataset, dist, loader_kwargs)

    # PaddedBatch as default collation for DynamicItemDataset
    if "collate_fn" not in loader_kwargs and isinstance(
        dataset, DynamicItemDataset
    ):
        loader_kwargs["collate_fn"] = PaddedBatch
    # Reproducible random sampling
    if loader_kwargs.get("shuffle", False):
        if loader_kwargs.get("sampler") is not None:
            raise ValueError(
                "Cannot specify both shuffle=True and a "
                "sampler in loader_kwargs"
            )
        sampler = ReproducibleRandomSampler(dataset)
        loader_kwargs["sampler"] = sampler
        # Should delete shuffle because you can't set both Sampler and
        # shuffle
        del loader_kwargs["shuffle"]
    # Create the loader
    if isinstance(dataset, IterableDataset):
        dataloader = DataLoader(dataset, **loader_kwargs)
    else:
        dataloader = SaveableDataLoader(dataset, **loader_kwargs)
    return dataloader


def train_loader_specifics(dataset, distributed_launch, loader_kwargs):
    rank = torch.distributed.get_rank() if torch.distributed.is_initialized() else 0
    sampler = loader_kwargs.get("sampler", None)
    shuffle = loader_kwargs.get("shuffle", False)
    if shuffle and not distributed_launch:
        if sampler is not None:
            raise ValueError(
                "Cannot specify both shuffle=True"
                "and a sampler in loader_kwargs"
            )
        sampler = ReproducibleRandomSampler(dataset)
        train_sampler = sampler
        loader_kwargs["sampler"] = train_sampler
        # Delete the shuffle flag, since you cannot specify both a sampler and
        # shuffling:
        del loader_kwargs["shuffle"]
    # Possibly make a DistributedSampler or a wrapper for some other sampler
    if distributed_launch and not isinstance(dataset, IterableDataset):
        drop_last = loader_kwargs.get("drop_last", False)
        # num_replicas arg is equal to world_size
        # and retrieved automatically within
        # DistributedSampler obj.
        if sampler is not None:
            train_sampler = DistributedSamplerWrapper(
                sampler,
                rank=rank,
                drop_last=drop_last,
                shuffle=shuffle,
            )
            # with DistributedSamplerWrapper, one must disable shuffling for dataloader
            loader_kwargs["shuffle"] = False
            loader_kwargs["sampler"] = train_sampler
        elif loader_kwargs.get("batch_sampler") is None:
            # no sampler and batch-sampler
            train_sampler = DistributedSampler(
                dataset, num_replicas=dist.get_world_size(), rank=rank, shuffle=False, drop_last=drop_last
            )
            # with DistributedSamplerWrapper, one must disable shuffling for dataloader
            loader_kwargs["shuffle"] = False
            loader_kwargs["sampler"] = train_sampler
        else:  # batch_sampler was specified
            train_sampler = DistributedSamplerWrapper(
                loader_kwargs.get("batch_sampler", None),
                rank=rank,
                shuffle=False,
            )
            loader_kwargs["batch_sampler"] = train_sampler
    elif distributed_launch and isinstance(dataset, IterableDataset):
        logger.warning(
            "Cannot automatically solve distributed sampling "
            "for IterableDataset."
        )
    return loader_kwargs

class SaveableDataLoader(DataLoader):
    """A saveable version of the PyTorch DataLoader.

    Note
    ----
    1. The saveability is implemented via some unfortunately slightly magical
    means.
    2. The data loader cannot recover after entering __iter__. Normally this is
    not a problem, as recovery should happen before training begins.  However,
    just before evaluation, it is also typical to recover the checkpoint at
    which performance was the best. Thus, if a checkpoint is loaded after
    entering __iter__, we just assume it is for this reason. A warning is
    logged, but that is all.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(self.dataset, IterableDataset):
            logging.warning(
                "SaveableDataLoader cannot save the position in an "
                "IterableDataset. Save the position on the dataset itself."
            )
        self._speechbrain_recovery_skip_to = None
        self._speechbrain_iterator = None

    def __iter__(self):
        iterator = super().__iter__()
        self._speechbrain_iterator = iterator
        return iterator

    @mark_as_saver
    def _speechbrain_save(self, path):
        if isinstance(self.dataset, IterableDataset):
            logging.warning(
                "Warning again: a checkpoint was requested on "
                "SaveableDataLoader, but the dataset is an IterableDataset. "
                "Cannot save the position in an IterableDataset. Not raising "
                "an error; assuming that you know what you're doing."
            )
        if self._speechbrain_iterator is None:
            to_save = None
        else:
            to_save = self._speechbrain_iterator._num_yielded
        with open(path, "w") as fo:
            fo.write(str(to_save))

    @mark_as_loader
    def _speechbrain_load(self, path, end_of_epoch, device=None):
        del device  # Unused here
        if self._speechbrain_iterator is not None:
            logging.debug(
                "SaveableDataLoader was requested to load a "
                "checkpoint, but the DataLoader has already been "
                "iterated. The DataLoader file will be ignored. "
                "This is normal in evaluation, when a checkpoint is "
                "loaded just to retrieve the best model."
            )
            return
        if end_of_epoch:
            # Don't load at end of epoch, as we actually want to start a fresh
            # epoch iteration next.
            return
        with open(path) as fi:
            saved = fi.read()
            if saved == str(None):
                # Saved at a point where e.g. an iterator did not yet exist.
                return
            else:
                self._speechbrain_recovery_skip_to = int(saved)



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import torch
from torch import nn

from e2eAIOK.DeNas.module.nlp.Linear_super import LinearSuper as SuperLinear


logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s -   %(message)s',
                    datefmt='%m/%d/%Y %H:%M:%S',
                    level=logging.INFO)
logger = logging.getLogger(__name__)

class SuperBertPooler(nn.Module):
    def __init__(self, config):
        super(SuperBertPooler, self).__init__()
        self.dense = SuperLinear(config.hidden_size, config.hidden_size)
        self.activation = nn.Tanh()

    def set_sample_config(self, sample_hidden_dim):
        self.dense.set_sample_config(sample_hidden_dim, sample_hidden_dim)

    def calc_sampled_param_num(self):
        dense_numel = self.dense.calc_sampled_param_num()
        #logger.info('dense_numel: {}'.format(dense_numel))

        return dense_numel

    def forward(self, hidden_states):
        # We "pool" the model by simply taking the hidden state corresponding
        # to the first token.
        first_token_tensor = hidden_states[:, 0]
        pooled_output = self.dense(first_token_tensor)
        pooled_output = self.activation(pooled_output)
        return pooled_output



# coding=utf-8
# Copyright (c) 2022, Intel. and its affiliates.
# Copyright (c) 2019, Facebook, Inc. and its affiliates. All Rights Reserved
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyTorch optimization for BERT model."""

import math
import torch
from torch.optim import Optimizer
from torch.optim.optimizer import required
from torch.nn.utils import clip_grad_norm_
import logging

from torch.optim.lr_scheduler import LambdaLR

logger = logging.getLogger(__name__)


def get_linear_schedule_with_warmup(optimizer, num_warmup_steps, num_training_steps, last_epoch=-1):
    """ Create a schedule with a learning rate that decreases linearly after
    linearly increasing during a warmup period.
    """

    def lr_lambda(current_step):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        return max(
            0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps))
        )

    return LambdaLR(optimizer, lr_lambda, last_epoch)


class AdamW(Optimizer):
    """ Implements Adam algorithm with weight decay fix.

    Parameters:
        lr (float): learning rate. Default 1e-3.
        betas (tuple of 2 floats): Adams beta parameters (b1, b2). Default: (0.9, 0.999)
        eps (float): Adams epsilon. Default: 1e-6
        weight_decay (float): Weight decay. Default: 0.0
        correct_bias (bool): can be set to False to avoid correcting bias in Adam (e.g. like in Bert TF repository). Default True.
    """

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0.0, correct_bias=True):
        if lr < 0.0:
            raise ValueError("Invalid learning rate: {} - should be >= 0.0".format(lr))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter: {} - should be in [0.0, 1.0[".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter: {} - should be in [0.0, 1.0[".format(betas[1]))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {} - should be >= 0.0".format(eps))
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, correct_bias=correct_bias)
        super().__init__(params, defaults)

    def reset_optimizer_states(self):
        for group in self.param_groups:
            for p in group['params']:
                # Exponential moving average of gradient values
                self.state[p]['exp_avg'] = torch.zeros_like(p.data)
                # Exponential moving average of squared gradient values
                self.state[p]['exp_avg_sq'] = torch.zeros_like(p.data)

    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                state = self.state[p]
                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p.data)

                state["step"] += 1

                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError("Adam does not support sparse gradients, please consider SparseAdam instead")

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]
                # Decay the first and second moment running average coefficient
                # In-place operations to update the averages at the same time
                exp_avg.mul_(beta1).add_(1.0 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1.0 - beta2, grad, grad)
                denom = exp_avg_sq.sqrt().add_(group["eps"])

                step_size = group["lr"]
                if group["correct_bias"]:  # No bias correction for Bert
                    bias_correction1 = 1.0 - beta1 ** state["step"]
                    bias_correction2 = 1.0 - beta2 ** state["step"]
                    step_size = step_size * math.sqrt(bias_correction2) / bias_correction1

                p.data.addcdiv_(-step_size, exp_avg, denom)
                if group["weight_decay"] > 0.0:
                    p.data.add_(-group["lr"] * group["weight_decay"], p.data)

        return loss


def warmup_cosine(x, warmup=0.002):
    if x < warmup:
        return x/warmup
    x_ = (x - warmup) / (1 - warmup)  # progress after warmup -
    return 0.5 * (1. + math.cos(math.pi * x_))


def warmup_constant(x, warmup=0.002):
    """ Linearly increases learning rate over `warmup`*`t_total` (as provided to BertAdam) training steps.
        Learning rate is 1. afterwards. """
    if x < warmup:
        return x/warmup
    return 1.0


def warmup_linear(x, warmup=0.002):
    """ Specifies a triangular learning rate schedule where peak is reached at `warmup`*`t_total`-th (as provided to BertAdam) training step.
        After `t_total`-th training step, learning rate is zero. """
    if x < warmup:
        return x/warmup
    return max((x-1.)/(warmup-1.), 0)


SCHEDULES = {
    'warmup_cosine':   warmup_cosine,
    'warmup_constant': warmup_constant,
    'warmup_linear':   warmup_linear,
}


class BertAdam(Optimizer):
    def __init__(self, params, lr=required, warmup=-1, t_total=-1, schedule='warmup_linear',
                 b1=0.9, b2=0.999, e=1e-6, weight_decay=0.01,
                 max_grad_norm=1.0):
        if lr is not required and lr < 0.0:
            raise ValueError("Invalid learning rate: {} - should be >= 0.0".format(lr))
        if schedule not in SCHEDULES:
            raise ValueError("Invalid schedule parameter: {}".format(schedule))
        if not 0.0 <= warmup < 1.0 and not warmup == -1:
            raise ValueError("Invalid warmup: {} - should be in [0.0, 1.0[ or -1".format(warmup))
        if not 0.0 <= b1 < 1.0:
            raise ValueError("Invalid b1 parameter: {} - should be in [0.0, 1.0[".format(b1))
        if not 0.0 <= b2 < 1.0:
            raise ValueError("Invalid b2 parameter: {} - should be in [0.0, 1.0[".format(b2))
        if not e >= 0.0:
            raise ValueError("Invalid epsilon value: {} - should be >= 0.0".format(e))
        defaults = dict(lr=lr, schedule=schedule, warmup=warmup, t_total=t_total,
                        b1=b1, b2=b2, e=e, weight_decay=weight_decay,
                        max_grad_norm=max_grad_norm)
        super(BertAdam, self).__init__(params, defaults)

    def get_lr(self):
        lr = []
        for group in self.param_groups:
            for p in group['params']:
                state = self.state[p]
                # print('state[key]: {}'.format(state['step']))
                if len(state) == 0:
                    return [0]
                if group['t_total'] != -1:
                    schedule_fct = SCHEDULES[group['schedule']]
                    lr_scheduled = group['lr'] * schedule_fct(state['step']/group['t_total'], group['warmup'])
                else:
                    lr_scheduled = group['lr']
                lr.append(lr_scheduled)
        return lr

    def reset_optimizer_states(self):
        for group in self.param_groups:
            for p in group['params']:
                # Exponential moving average of gradient values
                self.state[p]['next_m'] = torch.zeros_like(p.data)
                # Exponential moving average of squared gradient values
                self.state[p]['next_v'] = torch.zeros_like(p.data)

    def set_step(self, global_step):
        for group in self.param_groups:
            for p in group['params']:
                state = self.state[p]
                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['next_m'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['next_v'] = torch.zeros_like(p.data)

                state['step'] = global_step

    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()

        warned_for_t_total = False

        for group in self.param_groups:
            for p in group['params']:
                state = self.state[p]
                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['next_m'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['next_v'] = torch.zeros_like(p.data)

                # print('state[\'step\']: {}'.format(state['step']))
                if 'step' not in state:
                    state['step'] = 0
                state['step'] += 1

                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError('Adam does not support sparse gradients, please consider SparseAdam instead')

                next_m, next_v = state['next_m'], state['next_v']
                beta1, beta2 = group['b1'], group['b2']

                # Add grad clipping
                if group['max_grad_norm'] > 0:
                    clip_grad_norm_(p, group['max_grad_norm'])

                # Decay the first and second moment running average coefficient
                # In-place operations to update the averages at the same time
                next_m.mul_(beta1).add_(1 - beta1, grad)
                next_v.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                update = next_m / (next_v.sqrt() + group['e'])

                # Just adding the square of the weights to the loss function is *not*
                # the correct way of using L2 regularization/weight decay with Adam,
                # since that will interact with the m and v parameters in strange ways.
                #
                # Instead we want to decay the weights in a manner that doesn't interact
                # with the m/v parameters. This is equivalent to adding the square
                # of the weights to the loss with plain (non-momentum) SGD.
                if group['weight_decay'] > 0.0:
                    update += group['weight_decay'] * p.data

                if group['t_total'] != -1:
                    schedule_fct = SCHEDULES[group['schedule']]
                    progress = state['step']/group['t_total']
                    lr_scheduled = group['lr'] * schedule_fct(progress, group['warmup'])
                    # warning for exceeding t_total (only active with warmup_linear
                    if group['schedule'] == "warmup_linear" and progress > 1. and not warned_for_t_total:
                        logger.warning(
                            "Training beyond specified 't_total' steps with schedule. Learning rate set. "
                            "Please set 't_total' correctly.")
                        warned_for_t_total = True
                    # end warning
                else:
                    lr_scheduled = group['lr']

                update_with_lr = lr_scheduled * update
                # logger.info('lr_scheduled: {}'.format(lr_scheduled))
                p.data.add_(-update_with_lr)

                # step_size = lr_scheduled * math.sqrt(bias_correction2) / bias_correction1
                # No bias correction
                # bias_correction1 = 1 - beta1 ** state['step']
                # bias_correction2 = 1 - beta2 ** state['step']

        return loss
